@@ -3,30 +3,15 @@
 # Released under the GNU Public License 3 (or higher, your choice)
 # See the file COPYING for details.
 
-"""
-:mod:`gromacs.config` -- Configuration for GromacsWrapper
+""":mod:`gromacs.config` -- Configuration for GromacsWrapper
 ==========================================================
 
 The config module provides configurable options for the whole package;
-It mostly serves to define which gromacs tools and other scripts are
-exposed in the :mod:`gromacs` package and where template files are
-located. The user can configure *GromacsWrapper* by
+It serves to define how to handle log files, set where template files are
+located and which gromacs tools are exposed in the :mod:`gromacs` package.
 
-1. editing the global configuration file ``~/.gromacswrapper.cfg``
-
-2. placing template files into directories under ``~/.gromacswrapper``
-   (:data:`gromacs.config.configdir`) which can be processed instead
-   of the files that come with *GromacsWrapper*
-
-In order to **set up a basic configuration file and the directories**
-a user should execute :func:`gromacs.config.setup` at least once. It
-will prepare the user configurable area in their home directory and it
-will generate a default global configuration file
-``~/.gromacswrapper.cfg`` (the name is defined in
-:data:`CONFIGNAME`)::
-
-  import gromacs
-  gromacs.config.setup()
+In order to set up a basic configuration file and the directories
+a user can execute :func:`gromacs.config.setup`.
 
 If the configuration file is edited then one can force a rereading of
 the new config file with :func:`gromacs.config.get_configuration`::
@@ -126,40 +111,105 @@ loglevel *INFO* (see the python `logging module`_ for details).
 Gromacs tools and scripts
 -------------------------
 
-``load_*`` variables are lists that contain instructions to other
-parts of the code which packages and scripts should be wrapped.
+Fundamentally, GromacsWrapper makes existing Gromacs tools
+(executables) available as functions. In order for this to work, these
+executables must be found in the environment of the Python process
+that runs GromacsWrapper, and the user must list all the tools that
+are to be made available.
 
-.. autodata:: load_tools
-.. autodata:: load_scripts
+Setting up the environment
+~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-:data:`load_tools` is populated from lists of executable names in the
-configuration file.
+The standard way to set up the Gromacs environment is to source ``GMXRC`` in
+the shell before running the Python process. ``GMXRC`` adjusts a number of
+environment variables (such as :envvar:`PATH` and :envvar:`LD_LIBRARY_PATH`)
+but also sets Gromacs-specific environment variables such as :envvar:`GMXBIN`,
+:envvar:`GMXDATA`, and many others::
 
-The tool groups are strings that contain white-space separated file
-names of Gromacs tools. These lists determine which tools are made
-available as classes in :mod:`gromacs.tools`.
+   source /usr/local/bin/GMXRC
 
-A typical Gromacs tools section of the config file looks like this::
+(where the path to ``GMXRC`` is often set differently to disntinguish different
+installed versions of Gromacs).
+
+Alternatively, GromacsWrapper can itself source a ``GMXRC`` file and set the
+environment with the :func:`set_gmxrc_environment` function. The path to a
+``GMXRC`` file can be set in the config file in the ``[Gromacs]`` section as
+
+  [Gromacs]
+
+  GMXRC = /usr/local/bin/GMXRC
+
+When GromacsWrapper starts up, it tries to set the environment using the
+``GMXRC`` defined in the config file. If this is left empty or is not in the
+file, nothing is being done.
+
+.. autofunction:: set_gmxrc_environment
+
+
+List of tools
+~~~~~~~~~~~~~
+
+The list of Gromacs tools can be specified in the config file in the
+``[Gromacs]`` section with the ``tools`` variable.
+
+The tool groups are a list of names that determines which tools are made
+available as classes in :mod:`gromacs.tools`. If not provided
+GromacsWrapper will first try to load Gromacs 5.x then Gromacs 4.x
+tools.
+
+If you choose to provide a list, the Gromacs tools section of the config
+file can be like this::
 
    [Gromacs]
    # Release of the Gromacs package to which information in this sections applies.
    release = 4.5.3
 
-   # tools contains the file names of all Gromacs tools for which classes are generated.
-   # Editing this list has only an effect when the package is reloaded.
+   # tools contains the file names of all Gromacs tools for which classes are
+   # generated. Editing this list has only an effect when the package is
+   # reloaded.
    # (Note that this example has a much shorter list than the actual default.)
    tools =
          editconf make_ndx grompp genion genbox
-         grompp pdb2gmx mdrun
+         grompp pdb2gmx mdrun mdrun_d
 
-   # Additional gromacs tools that should be made available.
-   extra =
-         g_count     g_flux
-         a_gridcalc  a_ri3Dc  g_ri3Dc
-
-   # which tool groups to make available as gromacs.NAME
+   # which tool groups to make available
    groups = tools extra
 
+For Gromacs 5.x and later (e.g., 2021) use a section like the following, where
+driver commands are supplied::
+
+   [Gromacs]
+   # Release of the Gromacs package to which information in this sections applies.
+   release = 5.0.5
+
+   # GMXRC contains the path for GMXRC file which will be loaded. If not
+   provided is expected that it was sourced as usual before importing this
+   library.
+   GMXRC = /usr/local/gromacs/bin/GMXRC
+
+   # tools contains the command names of all Gromacs tools for which classes are generated.
+   # Editing this list has only an effect when the package is reloaded.
+   # (Note that this example has a much shorter list than the actual default.)
+   tools = gmx gmx_d
+
+
+For example, on the commandline you would run ::
+
+   gmx grompp -f md.mdp -c system.gro -p topol.top -o md.tpr
+
+and within GromacsWrapper this would become ::
+
+   gromacs.grompp(f="md.mdp", c="system.gro", p="topol.top", o="md.tpr")
+
+
+.. Note:: Because of `changes in the Gromacs tool in 5.x`_,
+          GromacsWrapper scripts might break, even if the tool
+          names are still the same.
+
+.. _`changes in the Gromacs tool in 5.x`:
+   http://www.gromacs.org/Documentation/How-tos/Tool_Changes_for_5.0
+
+
 
 Location of template files
 --------------------------
@@ -171,57 +221,71 @@ completely transparent to the user.
 
 .. autodata:: qscriptdir
 .. autodata:: templatesdir
-.. autodata:: managerdir
 .. autodata:: templates
 .. autodata:: qscript_template
 
 """
-from __future__ import with_statement
+from __future__ import absolute_import, with_statement, print_function
 
-import os, errno
-from ConfigParser import SafeConfigParser
+import os
+import logging
+import re
+import subprocess
+import sys
+
+if sys.version_info[0] < 3:  # several differences for Python 2
+    from ConfigParser import SafeConfigParser as ConfigParser
+    from ConfigParser import NoSectionError, NoOptionError
+    # Define read_file to point to the (deprecated in Python 3) readfp
+    # in order to have consistent, non-deprecated syntax
+    ConfigParser.read_file = ConfigParser.readfp
+
+    # Implement the new `fallback` kwarg based on the Python 3.7 implementation
+    # https://github.com/python/cpython/blob/3.7/Lib/configparser.py#L804
+    # This should ensure backwards compatibility.
+    _cf_getbool = ConfigParser.getboolean
+    _UNSET = object()
+
+    def _getboolean(self, section, option, fallback=_UNSET, **kwargs):
+        """Return a boolean for the specified config option
+
+        If *fallback* is used, it will be returned if there if the
+        option is not specified anywhere (defaults, config file)."""
+        try:  # Try using the Python 2 function
+            return _cf_getbool(self, section, option, **kwargs)
+        except (NoSectionError, NoOptionError):
+            if fallback is _UNSET:
+                raise
+        return fallback  # If fallback is given, use that value
+    ConfigParser.getboolean = _getboolean
+else:
+    from configparser import ConfigParser
 
 from pkg_resources import resource_filename, resource_listdir
 
-import utilities
+from . import utilities
 
-# Defaults
-# --------
-# hard-coded package defaults
 
-#: name of the global configuration file.
+#: Default name of the global configuration file.
 CONFIGNAME = os.path.expanduser(os.path.join("~",".gromacswrapper.cfg"))
 
-#: Holds the default values for important file and directory locations.
-#:
-#: :data:`configdir`
-#:    Directory to store user templates and configurations.
-#:    The default value is ``~/.gromacswrapper``.
-#: :data:`qscriptdir`
-#:    Directory to store user supplied queuing system scripts as
-#:    used by :mod:`gromacs.qsub`.
-#:    The default value is ``~/.gromacswrapper/qscripts``.
-#: :data:`templatesdir`
-#:    Directory to store user supplied template files such as mdp files.
-#:    The default value is ``~/.gromacswrapper/templates``.
-#: :data:`managerdir`
-#:    Directory to store configuration files for different queuing system
-#:    managers as used in :mod:`gromacs.manager`.
-#:    The default value is ``~/.gromacswrapper/managers``.
+#: Default configuration directory in GromacsWrapper.
+default_configdir = os.path.expanduser(os.path.join("~",".gromacswrapper"))
+
+#: Initial defaults for directories, filenames, and logger options.
 defaults = {
-     'configdir': os.path.expanduser(os.path.join("~",".gromacswrapper")),
-     'logfilename': "gromacs.log",
-     'loglevel_console': 'INFO',
-     'loglevel_file': 'DEBUG',
+    'configdir':    default_configdir,
+    'qscriptdir':   os.path.join(default_configdir, 'qscripts'),
+    'templatesdir': os.path.join(default_configdir, 'templates'),
+    'logfilename': "gromacs.log",
+    'loglevel_console': 'INFO',
+    'loglevel_file': 'DEBUG',
 }
-defaults['qscriptdir'] = os.path.join(defaults['configdir'], 'qscripts')
-defaults['templatesdir'] = os.path.join(defaults['configdir'], 'templates')
-defaults['managerdir'] = os.path.join(defaults['configdir'], 'managers')
 
 
 # Logging
 # -------
-import logging
+
 logger = logging.getLogger("gromacs.config")
 
 #: File name for the log file; all gromacs command and many utility functions (e.g. in
@@ -239,9 +303,6 @@ loglevel_file = logging.getLevelName(defaults['loglevel_file'])
 # User-accessible configuration
 # -----------------------------
 
-# (written in a clumsy manner because of legacy code and because of
-# the way I currently generate the documentation)
-
 #: Directory to store user templates and rc files.
 #: The default value is ``~/.gromacswrapper``.
 configdir = defaults['configdir']
@@ -254,13 +315,8 @@ qscriptdir = defaults['qscriptdir']
 #: The default value is ``~/.gromacswrapper/templates``.
 templatesdir = defaults['templatesdir']
 
-#: Directory to store configuration files for remote queuing systems
-#: :class:`gromacs.qsub.Manager` instances.
-#: The default value is ``~/.gromacswrapper/managers``.
-managerdir = defaults['managerdir']
-
 #: List of all configuration directories.
-config_directories = [configdir, qscriptdir, templatesdir, managerdir]
+config_directories = [configdir, qscriptdir, templatesdir]
 
 
 #: Search path for user queuing scripts and templates. The internal package-supplied
@@ -284,7 +340,7 @@ def _generate_template_dict(dirname):
     by external code. All template filenames are stored in
     :data:`config.templates`.
     """
-    return dict((resource_basename(fn), resource_filename(__name__, dirname+'/'+fn))
+    return dict((resource_basename(fn), resource_filename(__name__, dirname +'/'+fn))
                 for fn in resource_listdir(__name__, dirname)
                 if not fn.endswith('~'))
 
@@ -408,11 +464,11 @@ def _get_template(t):
               else:
                    _t_found = True
          if not _t_found:            # 5) nothing else to try...
-              raise ValueError("Failed to locate the template file %(t)r." % vars())
+              raise ValueError("Failed to locate the template file {t!r}.".format(**vars()))
     return os.path.realpath(t)
 
 
-class GMXConfigParser(SafeConfigParser):
+class GMXConfigParser(ConfigParser, object):
      """Customized :class:`ConfigParser.SafeConfigParser`."""
      cfg_template = 'gromacswrapper.cfg'
 
@@ -432,20 +488,19 @@ class GMXConfigParser(SafeConfigParser):
 
           self.filename = kwargs.pop('filename', CONFIGNAME)
 
-          args = tuple([self] + list(args))
-          SafeConfigParser.__init__(*args, **kwargs)  # old style class ... grmbl
+          super(GMXConfigParser, self).__init__(*args, **kwargs)
           # defaults
-          self.set('DEFAULT', 'configdir', defaults['configdir'])
           self.set('DEFAULT', 'qscriptdir',
                   os.path.join("%(configdir)s", os.path.basename(defaults['qscriptdir'])))
           self.set('DEFAULT', 'templatesdir',
                   os.path.join("%(configdir)s", os.path.basename(defaults['templatesdir'])))
-          self.set('DEFAULT', 'managerdir',
-                  os.path.join("%(configdir)s", os.path.basename(defaults['managerdir'])))
           self.add_section('Gromacs')
-          self.set("Gromacs", "tools", "pdb2gmx editconf grompp genbox genion mdrun trjcat trjconv")
+          self.set("Gromacs", "release", "")
+          self.set("Gromacs", "GMXRC", "")
+          self.set("Gromacs", "tools", "")
           self.set("Gromacs", "extra", "")
           self.set("Gromacs", "groups", "tools")
+          self.set("Gromacs", "append_suffix", "yes")
           self.add_section('Logging')
           self.set('Logging', 'logfilename', defaults['logfilename'])
           self.set('Logging', 'loglevel_console', defaults['loglevel_console'])
@@ -453,7 +508,7 @@ class GMXConfigParser(SafeConfigParser):
 
           # bundled defaults (should be ok to use get_template())
           default_cfg = get_template(self.cfg_template)
-          self.readfp(open(default_cfg))
+          self.read_file(open(default_cfg))
 
           # defaults are overriden by existing user global cfg file
           self.read([self.filename])
@@ -484,16 +539,6 @@ class GMXConfigParser(SafeConfigParser):
           """Return option as an expanded path."""
           return os.path.expanduser(os.path.expandvars(self.get(section, option)))
 
-     def getlist(self, section, option, separator=None, sort=True):
-          """Return *option* as a a list of strings.
-
-          Split on white space or *separator*. Sort for *sort* = ``True``.
-          """
-          l = self.get(section, option).split(separator)
-          if sort:
-               l.sort()
-          return l
-
      def getLogLevel(self, section, option):
           """Return the textual representation of logging level 'option' or the number.
 
@@ -555,7 +600,7 @@ def setup(filename=CONFIGNAME):
                cfg.write(configfile)  # write the default file so that user can edit
                msg = "NOTE: GromacsWrapper created the configuration file \n\t%r\n" \
                      "      for you. Edit the file to customize the package." % filename
-               print msg
+               print(msg)
 
      # directories
      for d in config_directories:
@@ -565,99 +610,100 @@ def setup(filename=CONFIGNAME):
 def check_setup():
      """Check if templates directories are setup and issue a warning and help.
 
-     Returns ``True`` if all files and directories are found and
-     ``False`` otherwise.
+    Set the environment variable  :envvar:`GROMACSWRAPPER_SUPPRESS_SETUP_CHECK`
+    skip the check and make it always return ``True``
 
-     Setting the environment variable
-     :envvar:`GROMACSWRAPPER_SUPPRESS_SETUP_CHECK` to 'true' ('yes'
-     and '1' also work) silence this function and make it always return ``True``.
+    :return ``True`` if directories were found and ``False`` otherwise
 
      .. versionchanged:: 0.3.1
-        Uses :envvar:`GROMACSWRAPPER_SUPPRESS_SETUP_CHECK` to suppress output
+        Uses :envvar:`GROMACSWRAPPER_SUPPRESS_SETUP_CHECK` to suppress check
         (useful for scripts run on a server)
      """
-     if os.environ.get("GROMACSWRAPPER_SUPPRESS_SETUP_CHECK", "false").lower() in ("1", "true", "yes"):
-         return True
 
-     is_complete = True
-     if not os.path.exists(CONFIGNAME):
-          is_complete = False
-          print "NOTE: The global configuration file %(CONFIGNAME)r is missing." % globals()
-          print "      You can create a basic configuration file from within python:"
-          print "      >>> import gromacs"
-          print "      >>> gromacs.config.setup()"
+     if "GROMACSWRAPPER_SUPPRESS_SETUP_CHECK" in os.environ:
+         return True
 
      missing = [d for d in config_directories if not os.path.exists(d)]
      if len(missing) > 0:
-          is_complete = False
-          print "NOTE: Some configuration directories are not set up yet"
-          print "      %r" % missing
-          print "      You can create them with the command from within Python"
-          print "      >>> import gromacs"
-          print "      >>> gromacs.config.setup()"
-     return is_complete
+         print("NOTE: Some configuration directories are not set up yet: ")
+         print("\t{0!s}".format('\n\t'.join(missing)))
+         print("NOTE: You can create the configuration file and directories with:")
+         print("\t>>> import gromacs")
+         print("\t>>> gromacs.config.setup()")
+         return False
+     return True
+
+
+def set_gmxrc_environment(gmxrc):
+    """Set the environment from ``GMXRC`` provided in *gmxrc*.
+
+    Runs ``GMXRC`` in a subprocess and puts environment variables loaded by it
+    into this Python environment.
+
+    If *gmxrc* evaluates to ``False`` then nothing is done. If errors occur
+    then only a warning will be logged. Thus, it should be safe to just call
+    this function.
+    """
+    # only v5: 'GMXPREFIX', 'GROMACS_DIR'
+    envvars = ['GMXBIN', 'GMXLDLIB', 'GMXMAN', 'GMXDATA',
+               'LD_LIBRARY_PATH', 'MANPATH', 'PKG_CONFIG_PATH',
+               'PATH',
+               'GMXPREFIX', 'GROMACS_DIR']
+    # in order to keep empty values, add ___ sentinels around result
+    # (will be removed later)
+    cmdargs = ['bash', '-c', ". {0} && echo {1}".format(gmxrc,
+               ' '.join(['___${{{0}}}___'.format(v) for v in envvars]))]
+
+    if not gmxrc:
+        logger.debug("set_gmxrc_environment(): no GMXRC, nothing done.")
+        return
+
+    try:
+        out = subprocess.check_output(cmdargs)
+        out = out.strip().split()
+        for key, value in zip(envvars, out):
+            value = str(value.decode('ascii').replace('___', ''))  # remove sentinels
+            os.environ[key] = value
+            logger.debug("set_gmxrc_environment(): %s = %r", key, value)
+    except (subprocess.CalledProcessError, OSError):
+        logger.warning("Failed to automatically set the Gromacs environment"
+                       "from GMXRC=%r", gmxrc)
+
+
+def get_tool_names():
+    """ Get tool names from all configured groups.
+
+    :return: list of tool names
+    """
+    names = []
+    for group in cfg.get('Gromacs', 'groups').split():
+        names.extend(cfg.get('Gromacs', group).split())
+    return names
+
+
+def get_extra_tool_names():
+    """ Get tool names from all configured groups.
+
+    :return: list of tool names
+    """
+    return cfg.get('Gromacs', 'extra').split()
+
+
+RELEASE = None
+MAJOR_RELEASE = None
+
+if cfg.get('Gromacs', 'release'):
+    RELEASE = cfg.get('Gromacs', 'release')
+    MAJOR_RELEASE = RELEASE.split('.')[0]
+
+for name in get_tool_names():
+    match = re.match(r'(gmx[^:]*):.*', name)
+    if match:
+        driver = match.group(1)
+        raise ValueError("'%s' isn't a valid tool name anymore."
+                         " Replace it by '%s'.\n            See "
+                         "https://gromacswrapper.readthedocs.io/en/latest/"
+                         "configuration.html" % (name, match.group(1)))
+
 
 check_setup()
-
-
-
-# Gromacs tools
-# -------------
-
-#: Python list of all tool file names. Filled from values in the tool
-#: groups in the configuration file.
-load_tools = []
-for g in cfg.getlist('Gromacs', 'groups', sort=False):
-     load_tools.extend(cfg.getlist('Gromacs', g))
-del g
-
-
-# Adding additional 3rd party scripts
-# -----------------------------------
-
-# HACK-Alert: we're temporarily "installing" scripts from the egg  ... this is a hack:
-# Must extract because it is part of a zipped python egg;
-# see http://peak.telecommunity.com/DevCenter/PythonEggs#accessing-package-resources
-GridMAT_MD = resource_filename(__name__,'external/GridMAT-MD_v1.0.2/GridMAT-MD.pl')
-os.chmod(GridMAT_MD, 0755)
-
-
-#: 3rd party bundled analysis scripts and tools; this is a list of triplets of
-#:
-#:   (*script name/path*, *command name*, *doc string*)
-#:
-#: (See the source code for examples.)
-load_scripts = [
-    (GridMAT_MD,
-     'GridMAT_MD',
-     """GridMAT-MD: A Grid-based Membrane Analysis Tool for use with Molecular Dynamics.
-
-*This* ``GridMAT-MD`` is a patched version of the original
-``GridMAT-MD.pl`` v1.0.2, written by WJ Allen, JA Lemkul and DR
-Bevan. The original version is available from the `GridMAT-MD`_ home
-page,
-
-.. _`GridMAT-MD`: http://www.bevanlab.biochem.vt.edu/GridMAT-MD/index.html
-
-Please cite
-
-  W. J. Allen, J. A. Lemkul, and D. R. Bevan. (2009) "GridMAT-MD: A
-  Grid-based Membrane Analysis Tool for Use With Molecular Dynamics."
-  J. Comput. Chem. 30 (12): 1952-1958.
-
-when using this programme.
-
-Usage:
-
-.. class:: GridMAT_MD(config[,structure])
-
-:Arguments:
-   - *config* : See the original documentation for a description for the
-     configuration file.
-   - *structure* : A gro or pdb file that overrides the value for
-     *bilayer* in the configuration file.
-
-"""),
-    ]
-
-
