@@ -36,6 +36,7 @@ directories:
    (uses :func:`anyopen`).
 
 .. autofunction:: anyopen
+.. autofunction:: isstream
 .. autofunction:: realpath
 .. function:: in_dir(directory[,create=True])
 
@@ -49,6 +50,7 @@ directories:
 
 .. autofunction:: find_first
 .. autofunction:: withextsep
+.. autofunction:: which
 
 Functions that improve list processing and which do *not* treat
 strings as lists:
@@ -76,6 +78,7 @@ Miscellaneous functions:
 
 .. autofunction:: convert_aa_code
 .. autofunction:: autoconvert
+.. autofunction:: hasmethod
 
 Data
 ----
@@ -83,46 +86,47 @@ Data
 .. autodata:: amino_acid_codes
 
 """
-from __future__ import with_statement
+from __future__ import absolute_import, with_statement, division
 
 __docformat__ = "restructuredtext en"
 
+import six
+from six import string_types
+
+import sys
 import os
 import glob
 import fnmatch
+import itertools
 import re
 import warnings
 import errno
 import subprocess
 from contextlib import contextmanager
-import bz2, gzip
+import bz2
+import gzip
 import datetime
 
+import numpy
+
 import logging
 logger = logging.getLogger('gromacs.utilities')
 
-from gromacs import AutoCorrectionWarning
-
-
-def Property(func):
-    """Simple decorator wrapper to make full fledged properties.
-    See eg http://adam.gomaa.us/blog/2008/aug/11/the-python-property-builtin/
-    """
-    return property(**func())
+from .exceptions import AutoCorrectionWarning
 
 
 class AttributeDict(dict):
     """A dictionary with pythonic access to keys as attributes --- useful for interactive work."""
-    def __getattribute__(self,x):
+    def __getattribute__(self, x):
         try:
             return super(AttributeDict,self).__getattribute__(x)
         except AttributeError:
             return self[x]
-    def __setattr__(self,name,value):
+    def __setattr__(self, name, value):
         try:
-            super(AttributeDict,self).__setitem__(name, value)
+            super(AttributeDict, self).__setitem__(name, value)
         except KeyError:
-            super(AttributeDict,self).__setattr__(name, value)
+            super(AttributeDict, self).__setattr__(name, value)
 
     def __getstate__(self):
         return self
@@ -130,108 +134,259 @@ class AttributeDict(dict):
     def __setstate__(self, state):
         self.update(state)
 
+
 def autoconvert(s):
     """Convert input to a numerical type if possible.
 
     1. A non-string object is returned as it is
     2. Try conversion to int, float, str.
     """
-    if not type(s) is str:
+    if type(s) is not str:
         return s
     for converter in int, float, str:   # try them in increasing order of lenience
         try:
-            return converter(s)
-        except ValueError:
+            s = [converter(i) for i in s.split()]
+            if len(s) == 1:
+                return s[0]
+            else:
+                return numpy.array(s)
+        except (ValueError, AttributeError):
             pass
-    raise ValueError("Failed to autoconvert %r" % s)
+    raise ValueError("Failed to autoconvert {0!r}".format(s))
 
 @contextmanager
-def openany(datasource, mode='r', **kwargs):
-    """Open the datasource and close it when the context exits.
+def openany(datasource, mode='rt', reset=True):
+    """Context manager for :func:`anyopen`.
+
+    Open the `datasource` and close it when the context of the :keyword:`with`
+    statement exits.
+
+    `datasource` can be a filename or a stream (see :func:`isstream`). A stream
+    is reset to its start if possible (via :meth:`~io.IOBase.seek` or
+    :meth:`~cString.StringIO.reset`).
+
+    The advantage of this function is that very different input sources
+    ("streams") can be used for a "file", ranging from files on disk (including
+    compressed files) to open file objects to sockets and strings---as long as
+    they have a file-like interface.
 
     :Arguments:
-       *datasource*
-          a stream or a filename
-       *mode*
-          ``'r'`` opens for reading, ``'w'`` for writing ['r']
-       *kwargs*
-          additional keyword arguments that are passed through to the
-          actual handler; if these are not appropriate then an
-          exception will be raised by the handler
+      *datasource*
+           a file or a stream
+      *mode*
+           {'r', 'w'} (optional), open in r(ead) or w(rite) mode
+      *reset*
+           bool (optional) try to read (`mode` 'r') the stream from the
+           start [``True``]
+
+
+    **Example**
+
+    Open a gzipped file and process it line by line::
+
+        with openany("input.pdb.gz") as pdb:
+            for line in pdb:
+                if line.startswith('ATOM'):
+                    print(line)
+
+    Open a URL and read it::
+
+       import urllib2
+       with openany(urllib2.urlopen("https://www.mdanalysis.org/")) as html:
+           print(html.read())
+
+
+    .. SeeAlso::
+       :func:`anyopen`
 
     """
-    stream, filename = anyopen(datasource, mode=mode, **kwargs)
+    stream = anyopen(datasource, mode=mode, reset=reset)
     try:
         yield stream
     finally:
         stream.close()
 
-def anyopen(datasource, mode='r', **kwargs):
+
+# On python 3, we want to use bz2.open to open and uncompress bz2 files. That
+# function allows to specify the type of the uncompressed file (bytes ot text).
+# The function does not exist in python 2, thus we must use bz2.BZFile to
+# which we cannot tell if the uncompressed file contains bytes or text.
+# Therefore, on python 2 we use a proxy function that removes the type of the
+# uncompressed file from the `mode` argument.
+try:
+    bz2.open
+except AttributeError:
+    # We are on python 2 and bz2.open is not available
+    def bz2_open(filename, mode):
+        """Open and uncompress a BZ2 file"""
+        mode = mode.replace('t', '').replace('b', '')
+        return bz2.BZ2File(filename, mode)
+else:
+    # We are on python 3 so we can use bz2.open
+    bz2_open = bz2.open
+
+
+def anyopen(datasource, mode='rt', reset=True):
     """Open datasource (gzipped, bzipped, uncompressed) and return a stream.
 
+    `datasource` can be a filename or a stream (see :func:`isstream`). By
+    default, a stream is reset to its start if possible (via
+    :meth:`~io.IOBase.seek` or :meth:`~cString.StringIO.reset`).
+
+    If possible, the attribute ``stream.name`` is set to the filename or
+    "<stream>" if no filename could be associated with the *datasource*.
+
     :Arguments:
-       *datasource*
-          a stream or a filename
-       *mode*
-          ``'r'`` opens for reading, ``'w'`` for writing ['r']
-       *kwargs*
-          additional keyword arguments that are passed through to the
-          actual handler; if these are not appropriate then an
-          exception will be raised by the handler
+      *datasource*
+        a file (from :class:`file` or :func:`open`) or a stream (e.g. from
+        :func:`urllib2.urlopen` or :class:`cStringIO.StringIO`)
+      *mode*
+        {'r', 'w', 'a'} (optional),
+        Open in r(ead), w(rite) or a(ppen) mode. More complicated
+        modes ('r+', 'w+', ...) are not supported; only the first letter of
+        `mode` is used and thus any additional modifiers are silently ignored.
+      *reset*
+        bool (optional),
+        try to read (`mode` 'r') the stream from the start
+
+    :Returns:
+       file-like object
+
+
+    .. SeeAlso::
+       :func:`openany` to be used with the :keyword:`with` statement.
 
     """
-    handlers = {'bz2': bz2.BZ2File, 'gz': gzip.open, '': file}
+    handlers = {'bz2': bz2_open, 'gz': gzip.open, '': open}
 
     if mode.startswith('r'):
-        if hasattr(datasource,'next') or hasattr(datasource,'readline'):
+        if isstream(datasource):
             stream = datasource
-            filename = '(%s)' % stream.name  # maybe that does not always work?
+            try:
+                filename = str(stream.name)  # maybe that does not always work?
+            except AttributeError:
+                filename = "<stream>"
+            if reset:
+                try:
+                    stream.reset()
+                except (AttributeError, IOError):
+                    try:
+                        stream.seek(0)
+                    except (AttributeError, IOError):
+                        warnings.warn("Stream {0}: not guaranteed to be at the beginning."
+                                      "".format(filename),
+                                      category=StreamWarning)
         else:
             stream = None
             filename = datasource
-            for ext in ('bz2', 'gz', ''):   # file == '' should be last
+            for ext in ('bz2', 'gz', ''):  # file == '' should be last
                 openfunc = handlers[ext]
-                stream = _get_stream(datasource, openfunc, mode=mode, **kwargs)
-                if not stream is None:
+                stream = _get_stream(datasource, openfunc, mode=mode)
+                if stream is not None:
                     break
             if stream is None:
-                raise IOError("Cannot open %(filename)r in mode=%(mode)r." % vars())
-    elif mode.startswith('w'):
-        if hasattr(datasource, 'write'):
+                raise IOError(errno.EIO, "Cannot open file or stream in mode={mode!r}.".format(**vars()), repr(filename))
+    elif mode.startswith('w') or mode.startswith('a'):  # append 'a' not tested...
+        if isstream(datasource):
             stream = datasource
-            filename = '(%s)' % stream.name  # maybe that does not always work?
+            try:
+                filename = str(stream.name)  # maybe that does not always work?
+            except AttributeError:
+                filename = "<stream>"
         else:
             stream = None
             filename = datasource
             name, ext = os.path.splitext(filename)
-            if ext.startswith(os.path.extsep):
+            if ext.startswith('.'):
                 ext = ext[1:]
             if not ext in ('bz2', 'gz'):
-                ext = ''   # anything else but bz2 or gz is just a normal file
+                ext = ''  # anything else but bz2 or gz is just a normal file
             openfunc = handlers[ext]
-            stream = openfunc(datasource, mode=mode, **kwargs)
+            stream = openfunc(datasource, mode=mode)
             if stream is None:
-                raise IOError("Cannot open %(filename)r in mode=%(mode)r with type %(ext)r." % vars())
+                raise IOError(errno.EIO, "Cannot open file or stream in mode={mode!r}.".format(**vars()), repr(filename))
     else:
-        raise NotImplementedError("Sorry, mode=%(mode)r is not implemented for %(datasource)r" % vars())
+        raise NotImplementedError("Sorry, mode={mode!r} is not implemented for {datasource!r}".format(**vars()))
+    try:
+        stream.name = filename
+    except (AttributeError, TypeError):
+        pass  # can't set name (e.g. cStringIO.StringIO)
+    return stream
 
-    return stream, filename
 
-def _get_stream(filename, openfunction=file, mode='r'):
+def _get_stream(filename, openfunction=open, mode='r'):
+    """Return open stream if *filename* can be opened with *openfunction* or else ``None``."""
     try:
         stream = openfunction(filename, mode=mode)
-    except IOError:
+    except (IOError, OSError) as err:
+        # An exception might be raised due to two reasons, first the openfunction is unable to open the file, in this
+        # case we have to ignore the error and return None. Second is when openfunction can't open the file because
+        # either the file isn't there or the permissions don't allow access.
+        if errno.errorcode[err.errno] in ['ENOENT', 'EACCES']:
+            six.reraise(*sys.exc_info())
         return None
-
-    try:
-        stream.readline()
-        stream.close()
-        stream = openfunction(filename,'r')
-    except IOError:
-        stream.close()
-        stream = None
+    if mode.startswith('r'):
+        # additional check for reading (eg can we uncompress) --- is this needed?
+        try:
+            stream.readline()
+        except IOError:
+            stream.close()
+            stream = None
+        except:
+            stream.close()
+            raise
+        else:
+            stream.close()
+            stream = openfunction(filename, mode=mode)
     return stream
 
+def hasmethod(obj, m):
+    """Return ``True`` if object *obj* contains the method *m*.
+
+    .. versionadded:: 0.7.1
+    """
+    return hasattr(obj, m) and callable(getattr(obj, m))
+
+def isstream(obj):
+    """Detect if `obj` is a stream.
+
+    We consider anything a stream that has the methods
+
+    - ``close()``
+
+    and either set of the following
+
+    - ``read()``, ``readline()``, ``readlines()``
+    - ``write()``, ``writeline()``, ``writelines()``
+
+    :Arguments:
+      *obj*
+          stream or str
+
+    :Returns:
+      *bool*, ``True`` if `obj` is a stream, ``False`` otherwise
+
+    .. SeeAlso::
+       :mod:`io`
+
+
+    .. versionadded:: 0.7.1
+    """
+    signature_methods = ("close",)
+    alternative_methods = (
+        ("read", "readline", "readlines"),
+        ("write", "writeline", "writelines"))
+
+    # Must have ALL the signature methods
+    for m in signature_methods:
+        if not hasmethod(obj, m):
+            return False
+    # Must have at least one complete set of alternative_methods
+    alternative_results = [
+        numpy.all([hasmethod(obj, m) for m in alternatives])
+        for alternatives in alternative_methods]
+    return numpy.any(alternative_results)
+
 # TODO: make it work for non-default charge state amino acids.
 #: translation table for 1-letter codes --> 3-letter codes
 #: .. Note: This does not work for HISB and non-default charge state aa!
@@ -240,7 +395,7 @@ amino_acid_codes = {'A':'ALA', 'C':'CYS', 'D':'ASP', 'E':'GLU',
                     'K':'LYS', 'L':'LEU', 'M':'MET', 'N':'ASN',
                     'P':'PRO', 'Q':'GLN', 'R':'ARG', 'S':'SER',
                     'T':'THR', 'V':'VAL', 'W':'TRP', 'Y':'TYR'}
-inverse_aa_codes = dict([(three, one) for one,three in amino_acid_codes.items()])
+inverse_aa_codes = {three: one for one,three in amino_acid_codes.items()}
 
 def convert_aa_code(x):
     """Converts between 3-letter and 1-letter amino acid codes."""
@@ -266,14 +421,14 @@ def in_dir(directory, create=True):
     try:
         try:
             os.chdir(directory)
-            logger.debug("Working in %(directory)r..." % vars())
-        except OSError, err:
+            logger.debug("Working in {directory!r}...".format(**vars()))
+        except OSError as err:
             if create and err.errno == errno.ENOENT:
                 os.makedirs(directory)
                 os.chdir(directory)
-                logger.info("Working in %(directory)r (newly created)..." % vars())
+                logger.info("Working in {directory!r} (newly created)...".format(**vars()))
             else:
-                logger.exception("Failed to start working in %(directory)r." % vars())
+                logger.exception("Failed to start working in {directory!r}.".format(**vars()))
                 raise
         yield os.getcwd()
     finally:
@@ -339,6 +494,28 @@ def find_files(directory, pattern):
                 filename = os.path.join(root, basename)
                 yield filename
 
+def which(program):
+    """Determine full path of executable *program* on :envvar:`PATH`.
+
+    (Jay at http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python)
+
+    .. versionadded:: 0.5.1
+    """
+
+    def is_exe(fpath):
+        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
+
+    fpath, fname = os.path.split(program)
+    if fpath:
+        real_program = realpath(program)
+        if is_exe(real_program):
+            return real_program
+    else:
+        for path in os.environ["PATH"].split(os.pathsep):
+            exe_file = os.path.join(path, program)
+            if is_exe(exe_file):
+                return exe_file
+    return None
 
 class FileUtils(object):
     """Mixin class to provide additional file-related capabilities."""
@@ -436,12 +613,12 @@ class FileUtils(object):
               ignored, do whatever *resolve* says
         """
         def _warn(x):
-            msg = "File %r already exists." % x
+            msg = "File {0!r} already exists.".format(x)
             logger.warn(msg)
             warnings.warn(msg)
             return True
         def _raise(x):
-            msg = "File %r already exists." % x
+            msg = "File {0!r} already exists.".format(x)
             logger.error(msg)
             raise IOError(errno.EEXIST, x, msg)
         solutions = {'ignore': lambda x: False,      # file exists, but we pretend that it doesn't
@@ -474,7 +651,7 @@ class FileUtils(object):
         return name
 
     def __repr__(self):
-        fmt = "%s(filename=%%r)" % self.__class__.__name__
+        fmt = "{0!s}(filename=%r)".format(self.__class__.__name__)
         try:
             fn =  self.filename()
         except ValueError:
@@ -484,7 +661,7 @@ class FileUtils(object):
 
 def iterable(obj):
     """Returns ``True`` if *obj* can be iterated over and is *not* a  string."""
-    if isinstance(obj, basestring):
+    if isinstance(obj, string_types):
         return False    # avoid iterating over characters of a string
     if hasattr(obj, 'next'):
         return True    # any iterator will do
@@ -513,7 +690,7 @@ def unlink_f(path):
     """Unlink path but do not complain if file does not exist."""
     try:
         os.unlink(path)
-    except OSError, err:
+    except OSError as err:
         if err.errno != errno.ENOENT:
             raise
 
@@ -538,7 +715,7 @@ def mkdir_p(path):
     """
     try:
         os.makedirs(path)
-    except OSError, err:
+    except OSError as err:
         if err.errno != errno.EEXIST:
             raise
 
@@ -549,11 +726,11 @@ def cat(f=None, o=None):
         return
     target = o
     infiles = asiterable(f)
-    logger.debug("cat %s > %s " % (" ".join(infiles), target))
+    logger.debug("cat {0!s} > {1!s} ".format(" ".join(infiles), target))
     with open(target, 'w') as out:
         rc = subprocess.call(['cat'] + infiles, stdout=out)
     if rc != 0:
-        msg = "failed with return code %d: cat %r > %r " % (rc, " ".join(infiles), target)
+        msg = "failed with return code {0:d}: cat {1!r} > {2!r} ".format(rc, " ".join(infiles), target)
         logger.exception(msg)
         raise OSError(errno.EIO, msg, target)
 
@@ -595,12 +772,12 @@ class Timedelta(datetime.timedelta):
     @property
     def dhours(self):
         """Hours component of the timedelta."""
-        return self.seconds / 3600
+        return self.seconds // 3600
 
     @property
     def dminutes(self):
         """Minutes component of the timedelta."""
-        return self.seconds/60 - 60*self.dhours
+        return self.seconds // 60 - 60*self.dhours
 
     @property
     def dseconds(self):
@@ -628,10 +805,10 @@ class Timedelta(datetime.timedelta):
         """
         substitutions = {
             "%d": str(self.days),
-            "%H": "%02d" % self.dhours,
+            "%H": "{0:02d}".format(self.dhours),
             "%h": str(24*self.days + self.dhours),
-            "%M": "%02d" % self.dminutes,
-            "%S": "%02d" % self.dseconds,
+            "%M": "{0:02d}".format(self.dminutes),
+            "%S": "{0:02d}".format(self.dseconds),
             }
         s = fmt
         for search, replacement in substitutions.items():
@@ -651,10 +828,8 @@ def number_pdbs(*args, **kwargs):
 
     format = kwargs.pop('format', "%(num)04d")
     name_format = "%(prefix)s" + format +".%(suffix)s"
-    filenames = []
-    map(filenames.append, map(glob.glob, args))  # concatenate all filename lists
-    filenames = filenames[0]                     # ... ugly
-    for f in filenames:
+
+    for f in itertools.chain.from_iterable(map(glob.glob, args)):
         m = NUMBERED_PDB.search(f)
         if m is None:
             continue
@@ -662,7 +837,7 @@ def number_pdbs(*args, **kwargs):
         prefix = m.group('PREFIX')
         suffix = m.group('SUFFIX')
         newname = name_format % vars()
-        logger.info("Renaming %(f)r --> %(newname)r" % vars())
+        logger.info("Renaming {f!r} --> {newname!r}".format(**vars()))
         try:
             os.rename(f, newname)
         except OSError:
